/*+++
 *  title:      MouseWheel.m
 *  abstract:   Implementation of MouseWheel InputManager for (OS|YB)/Windows.
 *  author:     Tom Hageman <tom@basil.icce.rug.nl>
 *  created:    January 1999, from an original hacque by Jeremy Bettis. 
 *  modified:   
 *  copyleft:
 *
 *		Copyright (C) 1999  Tom R. Hageman.
 *
 *	This is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this software; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 59 Temple Place - Suite 330,
 *	Boston, MA 02111-1307, USA.
 *
 *  description:
 *      (see corresponding *.h file)
 *---*/

#import "MouseWheel.h"

#define MOUSEWHEELSTEP_MIN	0.04	// Arbitrary lower limit for MouseWheelStep.


#ifdef WIN32
#import <windows.h>
#endif

#import <AppKit/NSApplication.h>
#import <AppKit/NSScrollView.h>
#import <AppKit/NSClipView.h>


#ifdef WIN32
// Some constants are missing from winuser.h, the NeXT header is too old
#define WM_MOUSEWHEEL           0x020A
#define WHEEL_DELTA             120     /* Value for rolling one detent */
#define WHEEL_PAGESCROLL        (UINT_MAX) /* Scroll one page */
#define MOUSEEVENTF_WHEEL       0x0800 /* wheel button rolled */


@interface _NSWin32EventHandler : NSObject
- (void)handleMachMessage:(void*)msg;
@end


@interface MouseWheel_NSWin32EventHandler : _NSWin32EventHandler
+ (void)_enableMouseWheel;
@end


@implementation MouseWheel_NSWin32EventHandler

+ (void)_enableMouseWheel
{
    if (self != [_NSWin32EventHandler class])
    {
        [self poseAsClass:[_NSWin32EventHandler class]];
    }
}

- (void)handleMachMessage:(void*)msg
{
    MSG event;
    if (PeekMessage(&event, 0, WM_MOUSEWHEEL, WM_MOUSEWHEEL, PM_REMOVE)) {
        /* This is a quick hack, of course.
	   In real life, we should deliver (and handle) it as an NSEvent. */
        if ( (short)HIWORD(event.wParam) > 0 ) {
            short distance;
            for (distance = HIWORD(event.wParam); distance > 0 ;
                distance -= WHEEL_DELTA)
                [NSApp sendAction:@selector(mouseWheeledUp:) to:0 from:0];
        } else if ( (short)HIWORD(event.wParam) < 0 ) {
            short distance;
            for (distance = HIWORD(event.wParam); distance < 0 ;
                distance += WHEEL_DELTA)
                [NSApp sendAction:@selector(mouseWheeledDown:) to:0 from:0];
        }
    } else
        [super handleMachMessage:msg];
}
@end
#endif // WIN32


@implementation NSScrollView (MouseWheelEvents)

- (float)mouseWheelDistance
{
    float line = [self lineScroll];
    float page = [self documentVisibleRect].size.height - [self pageScroll];
    float step = [MouseWheel mouseWheelStep];

    if (step >= 1)
    {
        // assume step = number of lines to scroll, but never more than a page.
        if (step >= (int)(page / line))
        {
            step = (int)(page / line) - 1;
            if (step <= 0) step = 1;
        }
        return (int)step * line;
    }

    // 0 < step < 1, assume step = fraction of page to scroll.
    page = page * step;
    if (line > page)
        return line;
    return (int)(page / line + 0.5) * line;
}

- (void)mouseWheeledUp:sender
{
    NSClipView * clip;
    NSPoint point;

    if (![self hasVerticalScroller])
        return;

    clip = [self contentView];
    point = [self documentVisibleRect].origin;
    if ([clip isFlipped]) {
        point.y -= [self mouseWheelDistance];
    } else {
        point.y += [self mouseWheelDistance];
    }
    point = [clip constrainScrollPoint:point];
    [clip scrollToPoint:point];
    [self reflectScrolledClipView:clip];
}

- (void) mouseWheeledDown:sender
{
    NSClipView * clip;
    NSPoint point;

    if (![self hasVerticalScroller])
        return;

    clip = [self contentView];
    point = [self documentVisibleRect].origin;
    if ([clip isFlipped]) {
        point.y += [self mouseWheelDistance];
    } else {
        point.y -= [self mouseWheelDistance];
    }
    point = [clip constrainScrollPoint:point];
    [clip scrollToPoint:point];
    [self reflectScrolledClipView:clip];
}

@end // NSScrollView (MouseWheelEvents)


@implementation MouseWheel

+ (void)load
{
    [self class]; // Causes +initialize to fire, which does additional setup.
}

+ (void)initialize
{
    if (self != [MouseWheel class]) return;

    [[NSUserDefaults standardUserDefaults] registerDefaults:
        [[[NSBundle bundleForClass:self] infoDictionary]
	    objectForKey:@"MouseWheelDefaults"]];

    [self setMouseWheelStep:[self defaultMouseWheelStep]];

    if ([[NSUserDefaults standardUserDefaults] boolForKey:@"MouseWheelEnable"])
    {
        [self enableMouseWheel];
    }
}

+ (void)enableMouseWheel
{
#ifdef WIN32
    [MouseWheel_NSWin32EventHandler _enableMouseWheel]; 
#endif
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(defaultsDidChange:) name:NSUserDefaultsDidChangeNotification object:nil];
}

+ (void)defaultsDidChange:(NSNotification *)aNotification
{
    [self setMouseWheelStep:[self defaultMouseWheelStep]];
}

+ (float)defaultMouseWheelStep
{
    return [[NSUserDefaults standardUserDefaults] floatForKey:@"MouseWheelStep"];
}

static float _mouseWheelStep;

+ (float)mouseWheelStep 
{
     return _mouseWheelStep;
}

+ (void)setMouseWheelStep:(float)aValue
{
    if (aValue < MOUSEWHEELSTEP_MIN) aValue = MOUSEWHEELSTEP_MIN;
    _mouseWheelStep = aValue;
}

@end // MouseWheel
